import numpy as np


class Math:
    """
    Class to simplify and encapsulate functionality related
    to math that could be more complex than the contained 
    in the basic python math module.
    """
    @staticmethod
    def sigmoid(value):
        """
        TODO: Write doc about this
        """
        return 1.0 / (1 + np.exp(-value))


from yta_general_utils.programming.error_message import ErrorMessage
# from yta_general_utils.file.filename import filename_is_type
# from yta_general_utils.checker.type import variable_is_positive_number
# from yta_general_utils.file.checker import file_exists
# from yta_general_utils.checker.url import url_is_ok
from enum import Enum
from typing import Union

import numpy as np


# TODO: Remove commented methods in PythonValidator for the next version
class PythonValidator:
    """
    Class to simplify and encapsulate the functionality related with
    parameters and variables validation.

    This class has been created to simplify the way it work and 
    replace the old ParameterValidator that was using too many 
    different methods being imported and generating cyclic import
    issues.

    We have some equivalent methods that do not need to pass the class
    as a parameter, so we can only work with the class name and avoid
    imports that can cause issues.
    """
    @staticmethod
    def is_instance(element, cls: Union[str, type]):
        """
        Check if the provided 'element' is an instance of the provided
        class 'cls'. An instance is not the same as a class. The 'cls'
        parameter can be the class or the string name of that class. 

        You can pass the string name of the class to avoid import in 
        the file where you are calling this method.
        """
        # TODO: Convert this to accept cls as array and check 
        # diferent classes
        if not PythonValidator.is_string(cls) and not PythonValidator.is_a_class(cls):
            raise Exception(f'The provided "cls" parameter "{str(cls)}" is not a string nor a class.')
        
        if PythonValidator.is_a_class(cls):
            return isinstance(element, cls)
        else:
            if PythonValidator.is_a_class(element):
                return False
        
            return getattr(type(element), '__name__', None) is cls
    
    @staticmethod
    def is_an_instance(element):
        """
        Check if the provided 'element' is an instance of any class.
        """
        return isinstance(element, object)
        # This below is just another alternative I want to keep
        return getattr(element, '__name__', None) is None
    
    @staticmethod
    def is_class(element, cls: Union[str, type]):
        """
        Check if the provided 'element' is the provided class 'cls'.
        A class is not the same as an instance of that class. The
        'cls' parameter can be the class or the string name of that
        class. 

        You can pass the string name of the class to avoid import in 
        the file where you are calling this method.
        """
        # TODO: Convert this to accept cls as array and check 
        # diferent classes
        if not PythonValidator.is_string(cls) and not PythonValidator.is_a_class(cls):
            raise Exception(f'The provided "cls" parameter "{str(cls)}" is not a string nor a class.')
        
        if PythonValidator.is_string(cls):
            return getattr(element, '__name__', None) is cls
        else:
            return PythonValidator.is_a_class(element) and element == cls

    @staticmethod
    def is_a_class(element):
        """
        Check if the provided 'element' is a class.
        """
        return isinstance(element, type)
        # This below is just another alternative I want to keep
        return getattr(element, '__name__', None) is not None
    
    @staticmethod
    def is_subclass(element: type, cls: Union[str, type]):
        """
        Check if the provided 'element' is a subclass of the provided
        class 'cls'. The 'cls' parameter can be the class or the
        string name of that class.

        You can pass the string name of the class to avoid import in 
        the file where you are calling this method.

        This method can return True if the provided 'element' is an
        instance or a class that inherits from Enum or YTAEnum class.
        Consider this above to check later if your 'element' is an
        instance or a class.
        """
        # TODO: Convert this to accept cls as array and check 
        # diferent classes
        if not PythonValidator.is_string(cls) and not PythonValidator.is_a_class(cls):
            raise Exception(f'The provided "cls" parameter "{str(cls)}" is not a string nor a class.')
        
        if not PythonValidator.is_a_class(element):
            # We want to know if it is a subclass, we don't
            # care if the provided one is an instance or a class,
            # so we try with both.
            if PythonValidator.is_an_instance(element):
                element = type(element)
            else:
                return False
            
        if PythonValidator.is_string(cls):
            return cls in [base_class.__name__ for base_class in element.__bases__]
        else: 
            issubclass(element, cls)
    
    @staticmethod
    def is_list(element):
        """
        Check if the provided 'element' is a list.
        """
        return type(element) == list
    
    @staticmethod
    def is_string(element):
        """
        Check if the provided 'element' is a string (str).
        """
        return isinstance(element, str)
    
    @staticmethod
    def is_boolean(element):
        """
        Check if the provided 'element' is a boolean (bool).
        """
        return isinstance(element, bool)
    
    @staticmethod
    def is_number(element, do_accept_string_number: bool = False):
        """
        Check if the provided 'element' is a numeric value. If
        'do_accept_string_number' is True, it will try to parse
        the 'element' as a float if a string is provided.
        """
        return NumberValidator.is_number(element, do_accept_string_number)

    @staticmethod
    def is_numpy_array(element):
        """
        Check if the provided 'element' is an instance of the
        numpy array np.ndarray.
        """
        return PythonValidator.is_instance(element, np.ndarray)
    
    @staticmethod
    def is_enum(element: Union['YTAEnum', Enum]):
        """
        Check if the provided 'element' is a subclass of an Enum or
        a YTAEnum.

        This method can return True if the provided 'element' is an
        instance or a class that inherits from Enum or YTAEnum class.
        Consider this above to check later if your 'element' is an
        instance or a class.
        """
        # TODO: I think it is 'EnumMeta' not Enum
        return PythonValidator.is_subclass(element, 'YTAEnum') or PythonValidator.is_subclass(element, 'Enum')
    
    @staticmethod
    def is_enum_instance(element: Union['YTAEnum', Enum]):
        """
        Check if the provided 'element' is a Enum (it is a subclass
        of an Enum or a YTAEnum) and it is an instance.
        """
        return PythonValidator.is_enum(element) and PythonValidator.is_an_instance(element)
    
    @staticmethod
    def is_enum_class(element: Union['YTAEnum', Enum]):
        """
        Check if the provided 'element' is a Enum (it is a subclass
        of an Enum or a YTAEnum) and it is a class.
        """
        return PythonValidator.is_enum(element) and PythonValidator.is_a_class(element)

class NumberValidator:
    """
    Class to simplify and encapsulate the functionality related
    to validate numeric values.
    """
    @staticmethod
    def is_number(element: Union[int, float, str], do_accept_string_number: bool = False):
        """
        Check if the provided 'element' is a numeric value. If
        'do_accept_string_number' is True, it will try to parse
        the 'element' as a float if a string is provided.
        """
        if not PythonValidator.is_instance(element, int) and not PythonValidator.is_instance(element, float) and not PythonValidator.is_instance(element, str):
            return False
        
        if PythonValidator.is_instance(element, str):
            if do_accept_string_number:
                try:
                    float(element)
                except:
                    return False
            else:
                return False
            
        return True
    
    @staticmethod
    def is_positive_number(element: Union[int, float, str], do_include_zero: bool = True):
        """
        This method checks if the provided 'element' is a numeric type,
        or tries to cast it as a float number if string provided, and
        returns True in the only case that the 'element' is actual a
        number by itself or as a string and it is 0 or above it. If 
        'do_include_zero' is set to False it won't be included.
        """        
        if not NumberValidator.is_number(element):
            return False
            
        variable = float(variable)

        if do_include_zero:
            return variable >= 0

        return variable > 0
    
    @staticmethod
    def is_number_between(element: Union[int, float, str], lower_limit: Union[int, float, str], upper_limit: Union[int, float, str], do_include_lower_limit: bool = True, do_include_upper_limit: bool = True):
        """
        This methods returns True if the provided 'variable' is a valid number
        that is between the also provided 'lower_limit' and 'upper_limit'. It
        will return False in any other case.
        """
        if not NumberValidator.is_number(element, True):
            return False
        
        if not NumberValidator.is_number(lower_limit) or not NumberValidator.is_number(upper_limit):
            return False
        
        element = float(element)
        lower_limit = float(lower_limit)
        upper_limit = float(upper_limit)
        
        # TODO: Should we switch limits if unordered (?)
        # if upper_limit < lower_limit:
        #     raise Exception(f'The provided "upper_limit" parameter {str(upper_limit)} is lower than the "lower_limit" parameter {str(lower_limit)} provided.')

        if do_include_lower_limit and do_include_upper_limit:
            return lower_limit <= element <= upper_limit
        elif do_include_lower_limit:
            return lower_limit <= element < upper_limit
        elif do_include_upper_limit:
            return lower_limit < element <= upper_limit
        else:
            return lower_limit < element < upper_limit
    
class ParameterValidator:
    @staticmethod
    def validate_mandatory_parameter(name: str, value):
        """
        Validates if the provided 'value' parameter with the also
        provided 'name' has a value, raising an Exception if not.

        This method returns the provided 'value' if everything is
        ok.
        """
        if not value:
            raise Exception(ErrorMessage.parameter_not_provided(name))

        return value

    @staticmethod
    def validate_string_parameter(name: str, value):
        """
        Validates if the provided 'value' parameter with the also
        provided 'name' is a string value, raising an Exception if
        not.

        This method returns the provided 'value' if everything is
        ok.
        """
        if not PythonValidator.is_string(value):
            raise Exception(ErrorMessage.parameter_is_not_string(name))

        return value
    
    @staticmethod
    def validate_bool_parameter(name: str, value: bool):
        """
        Validates if the provided 'value' parameter with the also
        provided 'name' is a boolean value, raising and Exception 
        if not.

        This method returns the provided 'value' if everything is
        ok.
        """
        if not PythonValidator.is_boolean(value):
            raise Exception(ErrorMessage.parameter_is_not_boolean(name))

        return value
    
# class ParameterValidator:
#     @classmethod
#     def validate_mandatory_parameter(cls, name: str, value):
#         """
#         Validates if the provided 'value' parameter with the also
#         provided 'name' has a value, raising an Exception if not.

#         This method returns the provided 'value' if everything is
#         ok.
#         """
#         if not value:
#             raise Exception(ErrorMessage.parameter_not_provided(name))

#         return value
        
#     @classmethod
#     def validate_string_parameter(cls, name: str, value: str):
#         """
#         Validates if the provided 'value' parameter with the also
#         provided 'name' is a string value, raising an Exception if
#         not.

#         This method returns the provided 'value' if everything is
#         ok.
#         """
#         if not isinstance(value, str):
#             raise Exception(ErrorMessage.parameter_is_not_string(name))

#         return value

#     @classmethod
#     def validate_bool_parameter(cls, name: str, value: bool):
#         """
#         Validates if the provided 'value' parameter with the also
#         provided 'name' is a boolean value, raising and Exception 
#         if not.

#         This method returns the provided 'value' if everything is
#         ok.
#         """
#         if not isinstance(value, bool):
#             raise Exception(ErrorMessage.parameter_is_not_boolean(name))

#         return value
        
#     @classmethod
#     def validate_file_exists(cls, name: str, value: str):
#         """
#         Validates if the provided 'value' parameter with the also
#         provided 'name' is a file that actually exists, raising
#         an Exception if not.

#         This method returns the provided 'value' if everything is
#         ok.
#         """
#         if not file_exists(value):
#             raise Exception(ErrorMessage.parameter_is_file_that_doesnt_exist(name))

#         return value
        
#     @classmethod
#     def validate_filename_is_type(cls, name: str, value: str, file_type: 'FileType'):
#         """
#         Validates if the provided 'value' parameter with the also
#         provided 'name' is a filename of the given 'file_type',
#         raising an Exception if not.

#         This method returns the provided 'value' if everything is
#         ok.
#         """
#         if not filename_is_type(value, file_type):
#             raise Exception(ErrorMessage.parameter_is_not_file_of_file_type(name, file_type))

#         return value
        
#     @classmethod
#     def validate_url_is_ok(cls, name: str, value: str):
#         """
#         Validates if the provided 'value' parameter with the also
#         provided 'name' is a valid url (the url is accessible),
#         raising an Exception if not.

#         This method returns the provided 'value' if everything is
#         ok.
#         """
#         if not url_is_ok(value):
#             raise Exception(ErrorMessage.parameter_is_not_valid_url(name))

#         return value
        
#     @classmethod
#     def validate_positive_number(cls, name: str, value: Union[int, float]):
#         """
#         Validates if the provided 'value' parameter with the also
#         provided 'name' is a positive number (0 or greater),
#         raising an Exception if not.

#         This method returns the provided 'value' as it is if 
#         everything is ok.
#         """
#         if not variable_is_positive_number(value):
#             raise Exception(ErrorMessage.parameter_is_not_positive_number(name))

#         return value

#     @classmethod
#     def validate_is_class(cls, name: str, value, class_names: Union[list[str], str]):
#         """
#         Validates if the provided 'value' is one of the provided 'class_names'
#         by using the 'type(value).__name__' checking.

#         This method returns the 'value' as it is if everything is ok.
#         """
#         if isinstance(class_names, str):
#             class_names = [class_names]

#         if not type(value).__name__ in class_names:
#             raise Exception(ErrorMessage.parameter_is_not_class(name, class_names))
        
#         return value
        
#     # Complex ones below
#     @classmethod
#     def validate_string_mandatory_parameter(cls, name: str, value: str):
#         """
#         Validates if the provided 'value' is a valid and non
#         empty string.
#         """
#         cls.validate_mandatory_parameter(name, value)
#         cls.validate_string_parameter(name, value)

#         return value

#     @classmethod
#     def validate_numeric_positive_mandatory_parameter(cls, name: str, value: str):
#         """
#         Validates if the provided 'value' is a positive numeric
#         value.
#         """
#         cls.validate_mandatory_parameter(name, value)
#         cls.validate_positive_number(name, value)

#         return value
    
#     @classmethod
#     def validate_is_enum_class(cls, enum: Union['YTAEnum', Enum]):
#         """
#         Validates if the provided 'value' is a valid Enum
#         class or subclass.

#         This method will raise an Exception if the provided
#         'value' is not a valid Enum class or subclass, or
#         will return it as it is if yes.
#         """
#         if not isinstance(enum, Enum) and not issubclass(enum, Enum):
#             raise Exception(f'The parameter "{enum}" provided is not an Enum class or subclass.')
        
#         return enum

#     @classmethod
#     def validate_enum(cls, value: Union['YTAEnum', str], enum: 'YTAEnum'):
#         """
#         Validates if the provided 'value' value is a valid
#         Enum or Enum value of the also provided 'enum' class.

#         This method will raise an Exception if something is
#         wrong or will return the 'value' as an 'enum' Enum.
#         instance if everything is ok.
#         """
#         cls.validate_mandatory_parameter('value', value)
#         cls.validate_is_enum_class(enum)
#         cls.validate_is_class('value', value, [enum.__class__.__name__, 'str'])

#         return enum.to_enum(value)


"""
    TODO:

    I have a problem here. If I use this as a library, I cannot 
    manually handle the different handlers I want to use nor 
    add them in the 'handle_shortcodes' method. I need to find
    a way in which I can use this in a better way.

    Maybe this is not suitable for the utils library and I have
    to use it manually in the project by coding itself inside 
    that project.
"""
from shortcodes import Parser
from enum import Enum


shortcodes_found = []
shortcode_previous_word_indexes = []

def specific_shortcode_handler(pargs, kwargs, context):
    """
    This is an specific shortcode handler that will specify the
    shortcode name in 'obj.shortcode' and what do we need to
    maintain in 'obj' variable to be able to work later.

    This method will be called everytime an ['specific_shortcode' ...]
    is found in the text in which we are parsing shortcodes, if this
    handler has been activated.
    """
    obj = {
        'shortcode': 'specific_shortcode',
        'previous_word_index': -1,
    }

    # TODO: This should be like this below and moved to specific folder
    class Example(Enum):
        KEYWORDS = 'keywords'
        DURATION = 'duration'
        ALL = [KEYWORDS, DURATION]

    for field in Example.ALL.value:
        if field in kwargs:
            obj[field] = kwargs[field]
    
    # TODO: Other fields later (why and which ones?)
    if 'audio' in kwargs:
        obj['audio'] = kwargs['audio']

    shortcodes_found.append(obj)
    
    return '[specific_shortcode]'

def handle_shortcodes(text):
    """
    This method processes the provided 'text' and finds the accepted shortcodes (see list below).
    It will return an object that contains 'shortcodes_found' and 'text_without_shortcodes'. At
    the same time, the 'shortcodes_found' is an array of objects that contains 'shortcode', 
    'keywords' and 'previous_word_index' indexes.

    'shortcodes' field includes the shortcode name. 'keywords' field includes the keywords to 
    look for the item we want, and 'previous_word_index' includes the index in which this word
    appears, to use it with the transcription to obtain the moment in which we should include 
    the result of this shortcode item.
    """
    # Found here:https://www.dmulholl.com/dev/shortcodes.html
    # We clean arrays to avoid strange behaviours
    global shortcodes_found
    global shortcode_previous_word_indexes

    # We register our shortcode parsers and ignore unsupported (unregistered)
    parser = Parser(start = '[', end = ']', ignore_unknown = True)
    parser.register(specific_shortcode_handler, 'specific_shortcode')

    shortcodes_found = []
    shortcode_previous_word_indexes = []

    # We parse all shortcodes
    output = parser.parse(text, context = None)
    words = output.split(' ')

    # Now I look for the previous word index for each shortcode
    index = 0
    while index < len(words):
        word = words[index]
        # TODO: Pay attention to shortcodes with characters next to ']' ('],', '].', etc.)
        # It doesn't get into this 'if' condition, so it fails later when accessing the 
        # previous word index
        if word.startswith('[') and word.endswith(']'):
            # Shortcode found, store previous word index
            if index == 0:
                # If shortcode is at the begining, treat it like if the previous word
                # was the first one, that would be exactly the same as if the shortcode
                # is just after the first word
                shortcode_previous_word_indexes.append(index)
            else:
                shortcode_previous_word_indexes.append(index - 1)
                print('Recognized "' + word + '" shortcode for the "' + words[index - 1] + '" word')
            # I remove the shortcode from the words array to safe next indexes
            del words[index]
            # I need to check again this position, the next word will be here
        else:
            index += 1

    # I save that index next to our shortcodes list object
    for index, shortcode in enumerate(shortcodes_found):
        shortcode['previous_word_index'] = shortcode_previous_word_indexes[index]

    # We clean the output (without shortcodes) to return it
    output = ' '.join(words)

    # TODO: What about shortcodes that have not been detected?
    return {
        'shortcodes_found': shortcodes_found.copy(),
        'text_without_shortcodes': output,
    }

        


from yta_general_utils.programming.enum import YTAEnum as Enum


class ImageFileExtension(Enum):
    """
    Enum class to encapsulate all existing image file extensions.
    """
    PNG = 'png'
    """
    Portable Network Graphics
    """
    JPEG = 'jpeg'
    """
    Joint Photographic Experts Group
    """
    JPG = 'jpg'
    """
    Joint Photographic Experts Group
    """
    WEBP = 'webp'
    """
    Web Picture
    """
    BMP = 'bmp'
    """
    Bitmap Image File
    """
    GIF = 'gif'
    """
    Graphics Interchange Format
    """
    TIFF = 'tiff'
    """
    Tagged Image File
    """
    PSD = 'psd'
    """
    Photoshop Document
    """
    PDF = 'pdf'
    """
    Portable Document Format
    """
    EPS = 'eps'
    """
    Encapsulated Postcript
    """
    AI = 'ai'
    """
    Adobe ILlustrator Document
    """
    INDD = 'indd'
    """
    Adobe Indesign Document
    """
    RAW = 'raw'
    """
    Raw Image Formats
    """
    CDR = 'cdr'
    """
    Corel Draw
    """
    # TODO: Add more

    @classmethod
    def default(cls):
        return cls.PNG

class AudioFileExtension(Enum):
    """
    Enum class to encapsulate all existing audio file extensions.
    """
    WAV = 'wav'
    """
    Waveform Audio
    """
    MP3 = 'mp3'
    """
    MPEG Audio Layer 3.
    """
    M4A = 'm4a'
    """
    MPEG-4 Audio
    """
    FLAC = 'flac'
    """
    Free Lossless Audio Codec.
    """
    WMA = 'wma'
    """
    Windows Media Audio
    """
    AAC = 'aac'
    """
    Advanced Audio Coding
    """
    # TODO: Add more

    @classmethod
    def default(cls):
        return cls.WAV

class VideoFileExtension(Enum):
    """
    Enum class to encapsulate all existing video file extensions.
    """
    MOV = 'mov'
    """
    Apple video
    """
    MP4 = 'mp4'
    """
    MPEG-4
    """
    WEBM = 'webm'
    """
    Developed by Google, subgroup of the open and standard Matroska Video Container (MKV)
    """
    AVI = 'avi'
    """
    Audio Video Interleave
    """
    WMV = 'wmv'
    """
    Windows Media Video
    """
    AVCHD = 'avchd'
    """
    Advanced Video Coding High Definition
    """
    FVL = '.flv'
    """
    Flash Video
    """
    # TODO: Add more

    @classmethod
    def default(cls):
        return cls.MP4

# These classes above should be used by the ffmpeg_handler and other
# declarations I make in our app to be consistent and reuse the code

class FileType(Enum):
    """
    Enum that represents the different file types and the valid
    extensions we accept for those file types. This Enum is to
    be used when checking filenames parameter.

    For example, we will use this to make sure the filename they
    gave to us is a video file type if we are storing a video 
    file.
    """
    IMAGE = ImageFileExtension.get_all()
    AUDIO = AudioFileExtension.get_all()
    VIDEO = VideoFileExtension.get_all()


"""
This is an adaption of the rate functions found in manim
library for the moviepy library. I have to use lambda t
functions with the current frame time to be able to resize
or reposition a video, so I have adapted the manim rate
functions to be able to return a factor that, with specific
moviepy functions, will make a change with the factor that
has been calculated with the corresponding rate function.

You can see 'manim/utils/rate_functions.py'.

This is the way I have found to make it work and to be able
to build smoother animations. As manim docummentation says,
the rate functions have been inspired by the ones listed in
this web page: https://easings.net/
"""
from yta_general_utils.math import Math
from math import pow, sqrt

import numpy as np
import typing


class RateFunction:
    """
    Class to simplify and encapsulate the functionality related
    to normalized bezier and rate functions.

    For any case the 'n' value must be a normalized value 
    between 0 and 1 to work as expected. You can apply the 
    result to a variation you want to make in some of your
    functions.

    If you want to make a linear zoom of a 20%, you will use
    a progressive zoom by applying the formula '1 + 0.2 * 
    linear(t)' (where t is the normalized time moment of the
    zoom animation).

    This is based on the manim library and this web page 
    https://easings.net/.
    """
    @staticmethod
    def linear(n: float) -> float:
        return n
    
    @staticmethod
    def slow_into(n: float) -> float:
        np.sqrt(1 - (1 - n) * (1 - n))

    @staticmethod
    def smooth(n: float, inflection: float = 10.0) -> float:
        error = Math.sigmoid(-inflection / 2)

        return min(
            max((Math.sigmoid(inflection * (n - 0.5)) - error) / (1 - 2 * error), 0),
            1,
        )
    
    @staticmethod
    def smoothstep(n: float) -> float:
        """
        Implementation of the 1st order SmoothStep sigmoid function.
        The 1st derivative (speed) is zero at the endpoints.
        https://en.wikipedia.org/wiki/Smoothstep
        """
        return 0 if n <= 0 else 3 * n**2 - 2 * n**3 if n < 1 else 1
    
    @staticmethod
    def smootherstep(n: float) -> float:
        """
        Implementation of the 2nd order SmoothStep sigmoid function.
        The 1st and 2nd derivatives (speed and acceleration) are zero at the endpoints.
        https://en.wikipedia.org/wiki/Smoothstep
        """
        return 0 if n <= 0 else 6 * n**5 - 15 * n**4 + 10 * n**3 if n < 1 else 1
    
    @staticmethod
    def smoothererstep(n: float) -> float:
        """
        Implementation of the 3rd order SmoothStep sigmoid function.
        The 1st, 2nd and 3rd derivatives (speed, acceleration and jerk) are zero at the endpoints.
        https://en.wikipedia.org/wiki/Smoothstep
        """
        alpha = 0
        if 0 < n < 1:
            alpha = 35 * n**4 - 84 * n**5 + 70 * n**6 - 20 * n**7
        elif n >= 1:
            alpha = 1
        return alpha
    
    @staticmethod
    def rush_into(n: float, inflection: float = 10.0) -> float:
        return 2 * RateFunction.smooth(n / 2.0, inflection)

    @staticmethod
    def rush_from(n: float, inflection: float = 10.0) -> float:
        return 2 * RateFunction.smooth(n / 2.0 + 0.5, inflection) - 1
    
    @staticmethod
    def double_smooth(n: float) -> float:
        if n < 0.5:
            return 0.5 * RateFunction.smooth(2 * n)
        else:
            return 0.5 * (1 + RateFunction.smooth(2 * n - 1))
        
    @staticmethod
    def there_and_back(n: float, inflection: float = 10.0) -> float:
        if n < 0.5:
            n = 2 * n
        else:
            n = 2 * (1 - n)

        return RateFunction.smooth(n, inflection)
    
    @staticmethod
    def there_and_back_with_pause(n: float, pause_ratio: float = 1.0 / 3) -> float:
        a = 1.0 / pause_ratio
        if n < 0.5 - pause_ratio / 2:
            return RateFunction.smooth(a * n)
        elif n < 0.5 + pause_ratio / 2:
            return 1
        else:
            return RateFunction.smooth(a - a * n)
        
    @staticmethod
    def running_start(n: float, pull_factor: float = -0.5) -> typing.Iterable: # what is func return type?
        # TODO: This is being taken from manim, maybe copy (?)
        from manim.utils.bezier import bezier

        return bezier([0, 0, pull_factor, pull_factor, 1, 1, 1])(n)
    
    @staticmethod
    def wiggle(n: float, wiggles: float = 2) -> float:
        # TODO: Manim says 'not_quite_there' function is not working
        return RateFunction.there_and_back(n) * np.sin(wiggles * np.pi * n)
    
    @staticmethod
    def ease_in_cubic(n: float) -> float:
        return n * n * n
    
    @staticmethod
    def ease_out_cubic(n: float) -> float:
        return 1 - pow(1 - n, 3)
    
    @staticmethod
    def squish_rate_func(func: typing.Callable[[float], float], a: float = 0.4, b: float = 0.6) -> typing.Callable[[float], float]:
        def result(n: float):
            if a == b:
                return a

            if n < a:
                return func(0)
            elif n > b:
                return func(1)
            else:
                return func((n - a) / (b - a))

        return result
    
    @staticmethod
    def lingering(n: float) -> float:
        return RateFunction.squish_rate_func(lambda n: n, 0, 0.8)(n)
    
    @staticmethod
    def exponential_decay(n: float, half_life: float = 0.1) -> float:
        # The half-life should be rather small to minimize
        # the cut-off error at the end
        return 1 - np.exp(-n / half_life)

    @staticmethod
    def ease_in_sine(n: float) -> float:
        return 1 - np.cos((n * np.pi) / 2)
    
    @staticmethod
    def ease_out_sine(n: float) -> float:
        return np.sin((n * np.pi) / 2)
    
    @staticmethod
    def ease_in_out_sine(n: float) -> float:
        return -(np.cos(np.pi * n) - 1) / 2

    @staticmethod
    def ease_in_quad(n: float) -> float:
        return n * n
    
    @staticmethod
    def ease_out_quad(n: float) -> float:
        return 1 - (1 - n) * (1 - n)

    @staticmethod
    def ease_in_out_quad(n: float) -> float:
        return 2 * n * n if n < 0.5 else 1 - pow(-2 * n + 2, 2) / 2
    
    @staticmethod
    def ease_in_out_cubic(n: float) -> float:
        return 4 * n * n * n if n < 0.5 else 1 - pow(-2 * n + 2, 3) / 2

    @staticmethod
    def ease_in_quart(n: float) -> float:
        return n * n * n * n
    
    @staticmethod
    def ease_out_quart(n: float) -> float:
        return 1 - pow(1 - n, 4)

    @staticmethod
    def ease_in_out_quart(n: float) -> float:
        return 8 * n * n * n * n if n < 0.5 else 1 - pow(-2 * n + 2, 4) / 2
    
    @staticmethod
    def ease_in_quint(n: float) -> float:
        return n * n * n * n * n

    @staticmethod
    def ease_out_quint(n: float) -> float:
        return 1 - pow(1 - n, 5)
    
    @staticmethod
    def ease_in_out_quint(n: float) -> float:
        return 16 * n * n * n * n * n if n < 0.5 else 1 - pow(-2 * n + 2, 5) / 2

    @staticmethod
    def ease_in_expo(n: float) -> float:
        return 0 if n == 0 else pow(2, 10 * n - 10)
    
    @staticmethod
    def ease_out_expo(n: float) -> float:
        return 1 if n == 1 else 1 - pow(2, -10 * n)
    
    @staticmethod
    def ease_in_out_expo(n: float) -> float:
        if n == 0:
            return 0
        elif n == 1:
            return 1
        elif n < 0.5:
            return pow(2, 20 * n - 10) / 2
        else:
            return (2 - pow(2, -20 * n + 10)) / 2
        
    @staticmethod
    def ease_in_circ(n: float) -> float:
        return 1 - sqrt(1 - pow(n, 2))
    
    @staticmethod
    def ease_out_circ(n: float) -> float:
        return sqrt(1 - pow(n - 1, 2))
    
    @staticmethod
    def ease_in_out_circ(n: float) -> float:
        return (
            (1 - sqrt(1 - pow(2 * n, 2))) / 2
            if n < 0.5
            else (sqrt(1 - pow(-2 * n + 2, 2)) + 1) / 2
        )

    @staticmethod
    def ease_in_back(n: float) -> float:
        c1 = 1.70158
        c3 = c1 + 1

        return c3 * n * n * n - c1 * n * n

    @staticmethod
    def ease_out_back(n: float) -> float:
        c1 = 1.70158
        c3 = c1 + 1

        return 1 + c3 * pow(n - 1, 3) + c1 * pow(n - 1, 2)
    
    @staticmethod
    def ease_in_out_back(n: float) -> float:
        c1 = 1.70158
        c2 = c1 * 1.525

        return (
            (pow(2 * n, 2) * ((c2 + 1) * 2 * n - c2)) / 2
            if n < 0.5
            else (pow(2 * n - 2, 2) * ((c2 + 1) * (n * 2 - 2) + c2) + 2) / 2
        )
    
    @staticmethod
    def ease_in_elastic(n: float) -> float:
        c4 = (2 * np.pi) / 3
        if n == 0:
            return 0
        elif n == 1:
            return 1
        else:
            return -pow(2, 10 * n - 10) * np.sin((n * 10 - 10.75) * c4)
        
    @staticmethod
    def ease_out_elastic(n: float) -> float:
        c4 = (2 * np.pi) / 3
        if n == 0:
            return 0
        elif n == 1:
            return 1
        else:
            return pow(2, -10 * n) * np.sin((n * 10 - 0.75) * c4) + 1

    @staticmethod
    def ease_in_out_elastic(n: float) -> float:
        c5 = (2 * np.pi) / 4.5
        if n == 0:
            return 0
        elif n == 1:
            return 1
        elif n < 0.5:
            return -(pow(2, 20 * n - 10) * np.sin((20 * n - 11.125) * c5)) / 2
        else:
            return (pow(2, -20 * n + 10) * np.sin((20 * n - 11.125) * c5)) / 2 + 1

    @staticmethod
    def ease_in_bounce(n: float) -> float:
        return 1 - RateFunction.ease_out_bounce(1 - n)
    
    @staticmethod
    def ease_out_bounce(n: float) -> float:
        n1 = 7.5625
        d1 = 2.75

        if n < 1 / d1:
            return n1 * n * n
        elif n < 2 / d1:
            return n1 * (n - 1.5 / d1) * (n - 1.5 / d1) + 0.75
        elif n < 2.5 / d1:
            return n1 * (n - 2.25 / d1) * (n - 2.25 / d1) + 0.9375
        else:
            return n1 * (n - 2.625 / d1) * (n - 2.625 / d1) + 0.984375
        
    @staticmethod
    def ease_in_out_bounce(n: float) -> float:
        if n < 0.5:
            return (1 - RateFunction.ease_out_bounce(1 - 2 * n)) / 2
        else:
            return (1 + RateFunction.ease_out_bounce(2 * n - 1)) / 2

# TODO: I can create my own curves by setting nodes with
# different values (as speed curves in a famous video
# editor) to make my own animations


from typing import Union


class ErrorMessage:
    @staticmethod
    def parameter_is_not_a_class(parameter_name: str):
        return f'The provided "{parameter_name}" parameter is not a class.'
    
    # TODO: All this methods below must be 'staticmethod', we do not
    # use instances with this class and will not use them.
    @classmethod
    def parameter_not_provided(cls, parameter_name: str):
        return f'The parameter "{parameter_name}" was not provided.'
    
    @classmethod
    def parameter_is_not_string(cls, parameter_name: str):
        return f'The parameter "{parameter_name}" provided is not a string.'
    
    @classmethod
    def parameter_is_not_boolean(cls, parameter_name: str):
        return f'The "{parameter_name}" parameter is not boolean.'
    
    @classmethod
    def parameter_is_not_positive_number(cls, parameter_name: str):
        return f'The parameter "{parameter_name}" provided is not a valid and positive number.'
    
    @classmethod
    def parameter_is_file_that_doesnt_exist(cls, parameter_name: str):
        return f'The "{parameter_name}" parameter provided is not a file that exists.'
    
    @classmethod
    def parameter_is_not_file_of_file_type(cls, parameter_name: str, file_type: 'FileType'):
        return f'The "{parameter_name}" provided is not a {file_type.value} filename.'
    
    @classmethod
    def parameter_is_not_valid_url(cls, parameter_name: str):
        return f'The provided "{parameter_name}" parameter is not a valid url.'
    
    @classmethod
    def parameter_is_not_class(cls, parameter_name: str, class_names: Union[list[str], str]):
        # TODO: Check if 'class_names' is not array of str nor str
        if isinstance(class_names, str):
            class_names = [class_names]

        class_names = ', '.join(class_names)
        return f'The provided "{parameter_name}" parameter is not one of the next classes: {class_names}'

    @classmethod
    def parameter_is_not_name_of_ytaenum_class(cls, name: str, enum):
        return f'The provided YTAEnum name "{name}" is not a valid {enum.__class__.__name__} YTAEnum name.'
    
    @classmethod
    def parameter_is_not_value_of_ytaenum_class(cls, value: any, enum):
        return f'The provided YTAEnum value "{value}" is not a valid {enum.__class__.__name__} YTAEnum value.'
    
    @classmethod
    def parameter_is_not_name_nor_value_of_ytaenum_class(cls, name_or_value: any, enum):
        return f'The provided YTAEnum name or value "{name_or_value}" is not a valid {enum.__class__.__name__} YTAEnum name or value.'
    
    @classmethod
    def parameter_is_not_name_nor_value_nor_enum_of_ytaenum_class(cls, name_or_value_or_enum: any, enum):
        return f'The provided YTAEnum name, value or instance "{name_or_value_or_enum}" is not a valid {enum.__class__.__name__} YTAEnum name, value or instance.'


from yta_general_utils.file.writer import FileWriter
from typing import Union

import requests


def get_file(url: str, output_filename: Union[str, None] = None):
    """
    This method sends a request to the provided 'url'
    if provided and obtains the file content (if 
    possible). It will write the obtained file locally
    as 'output_filename' if provided.

    This method returns the file content data as 
    obtained from the requested (.content field).
    """
    if not url:
        raise Exception('No "url" provided.')
    
    content = requests.get(url).content

    if output_filename:
        FileWriter.write_binary_file(content, output_filename)

    return content

def download_file(url: str, output_filename: str):
    """
    Receives a downloadable url as 'url' and downloads that file in
    our system as 'output_filename'.
    """
    if not output_filename:
        raise Exception('No "output_filename" provided to save the file.')

    return get_file(url, output_filename)


# TODO: Find a more specific and better way of classifying these
# files and methods. I moved here to avoid using __init__.py

from yta_general_utils.file.filename import sanitize_filename
from yta_general_utils.file.checker import FileValidator
from yta_general_utils.programming.enum import YTAEnum as Enum
from shutil import move as shutil_move, copyfile as shutil_copyfile

import glob


class FileSearchOption(Enum):
    """
    Enum that allows us setting the strategy dynamically when 
    searching for files.
    """
    FILES_AND_FOLDERS = 'fifo'
    """
    This option, when set, will return files and folders.
    """
    FILES_ONLY = 'fi'
    """
    This option, when set, will return files only.
    """
    FOLDERS_ONLY = 'fo'
    """
    This option, when set, will return folders only.
    """

class FileHandler:
    @staticmethod
    def get_list(abspath, option: FileSearchOption = FileSearchOption.FILES_AND_FOLDERS, pattern: str = '*', recursive: bool = False):
        """
        List what is inside the provided 'abspath'. This method will list files and
        folders, files or only folders attending to the provided 'option'. It will
        also filter the files/folders that fit the provided 'pattern' (you can use
        '*' as wildcard, so for example '*.jpg' will list all images). This method
        can also be used in a recursive way if 'recursive' parameter is True, but
        take care of memory consumption and it would take its time to perform.

        This method returns a list with all existing elements absolute paths 
        sanitized.
        """
        if not abspath:
            return None
        
        abspath = sanitize_filename(abspath)
        list = []

        # This below get files and folders
        files_and_folders = [sanitize_filename(f) for f in glob.glob(pathname = abspath + pattern, recursive = recursive)]

        if option == FileSearchOption.FILES_ONLY:
            for f in files_and_folders:
                if FileValidator.is_file(f):
                    list.append(f)
        elif option == FileSearchOption.FOLDERS_ONLY:
            for f in files_and_folders:
                if FileValidator.is_folder(f):
                    list.append(f)
        elif option == FileSearchOption.FILES_AND_FOLDERS:
            list = files_and_folders
        
        return list

    @staticmethod
    def rename_file(origin_filename: str, destination_filename: str, replace_if_existing: bool = False):
        """
        Renames the 'origin_filename' to the 'destination_filename'.
        If 'replace_if_existing' is True, it will replace the destination
        file if possible and allowed. If it is False, it will fail.

        TODO: Remove 'replace_if_existing' if not used.
        """
        if not origin_filename:
            return None
        
        if not destination_filename:
            return None
        
        shutil_move(origin_filename, destination_filename)

    @staticmethod
    def copy_file(origin_filename: str, destination_filename: str):
        """
        Makes a copy of the provided 'origin_filename' and 
        stores it as 'destination_filename'.

        The destination folder must exist.
        """
        if not origin_filename:
            return None

        if not destination_filename:
            return None

        shutil_copyfile(origin_filename, destination_filename)


from yta_general_utils.image.checker import is_valid_image
from yta_general_utils.file.filename import sanitize_filename
from moviepy.editor import AudioFileClip, VideoFileClip
from pathlib import Path


class FileValidator:
    """
    Class to simplify and encapsulate the functionality related
    to file and folders validation.
    """
    @staticmethod
    def is_file(filename):
        """
        Checks if the provided 'filename' is an existing and
        valid file. It returns True if yes or False if not.
        """
        filename = sanitize_filename(filename)
        filename = Path(filename)

        try:
            return filename.exists() and filename.is_file()
        except:
            # TODO: Maybe print stack (?)
            return False

    @staticmethod
    def is_folder(filename):
        """
        Checks if the provided 'filename' is an existing and
        valid folder. It returns True if yes or False if not.
        """
        filename = sanitize_filename(filename)
        filename = Path(filename)

        try:
            return filename.exists() and filename.is_dir()
        except:
            # TODO: Maybe print stack (?)
            return False

    @staticmethod
    def file_exists(filename):
        """
        Checks if the provided 'filename' file or folder exist. It
        returns True if existing or False if not. This method
        sanitizes the provided 'filename' before checking it.
        """
        filename = sanitize_filename(filename)

        try:
            return Path(filename).exists()
        except:
            # TODO: Maybe print stack (?)
            return False

    @staticmethod
    def file_is_audio_file(filename):
        """
        Checks if the provided 'filename' is an audio file by
        trying to instantiate it as a moviepy AudioFileClip.
        This method sanitizes the provided 'filename' before 
        checking it.
        """
        filename = sanitize_filename(filename)

        try:
            AudioFileClip(filename)
        except:
            return False
        
        return True

    @staticmethod
    def file_is_video_file(filename):
        """
        Checks if the provided 'filename' is a video file by
        trying to instantiate it as a moviepy VideoFileClip.
        This method sanitizes the provided 'filename' before
        checking it.
        """
        filename = sanitize_filename(filename)

        try:
            VideoFileClip(filename)
        except:
            return False
        
        return True

    @staticmethod
    def file_is_image_file(filename: str):
        """
        Checks if the provided 'filename' is a valid image
        file by opening and verifying it with the Pillow
        library.
        """
        filename = sanitize_filename(filename)

        return is_valid_image(filename)


from yta_general_utils.temp import create_temp_filename
from yta_general_utils.image.processor import remove_background
from PIL import Image

import numpy as np
import cv2


# TODO: This is related to Image so multimedia (?)
# This methods has been placed in yta_multimedia, I 
# think, so remove them from here
def pixelate(image_filename, i_size, output_filename):
    """
    Pixelates the provided 'image_filename' and saves it as the 'output_filename'.
    The 'i_size' is the pixelating square. The smaller it is, the less pixelated 
    its.

    'i_size' must be a tuple such as (8, 8) or (16, 16).
    """
    #read file
    img = Image.open(image_filename)

    # convert to small image
    small_img = img.resize(i_size,Image.BILINEAR)

    # resize to output size
    res = small_img.resize(img.size, Image.NEAREST)

    res.save(output_filename)

def to_sticker(image_filename, output_filename = None):
    """
    Receives an image and turns it into an sticker. This method will remove the 
    background of the provided 'image_filename' and surrounds the main object
    in that picture with a wide white border (as social networks stickers). It 
    will also crop the image to fit the remaining object only.
    """
    # From here: https://withoutbg.com/resources/creating-sticker
    # We enlarge the image 40 pixels (each border) to ensure sticker works well
    # TODO: Move this to a custom method
    enlarged_filename = create_temp_filename('enlarged.png')
    old_im = Image.open(image_filename)
    old_size = old_im.size

    new_size = (old_size[0] + 80, old_size[1] + 80)
    new_im = Image.new("RGB", new_size)
    box = tuple((n - o) // 2 for n, o in zip(new_size, old_size))
    new_im.paste(old_im, box)
    new_im.save(enlarged_filename)

    # We remove the background of the new large image
    without_background_filename = create_temp_filename('without_background.png')
    remove_background(enlarged_filename, without_background_filename)

    img = __read_image(without_background_filename)
    alpha = __extract_alpha_channel(img)
    big_contour = __get_largest_contour(alpha)
    contour_img = __draw_filled_contour_on_black_background(big_contour, alpha.shape)
    dilate = __apply_dilation(contour_img)
    canvas = np.zeros(img.shape, dtype = np.uint8)
    canvas = __apply_overlays(canvas, img, dilate)
    # The image is as larg as the original, maybe we need to crop it
    large_result_filename = create_temp_filename('large.png')
    Image.fromarray(canvas.astype(np.uint8), mode = 'RGBA').save(large_result_filename)
    # Cropping considering non-alpha pixels as dimension to preserve
    im = cv2.imread(large_result_filename, cv2.IMREAD_UNCHANGED)
    x, y, w, h = cv2.boundingRect(im[..., 3])
    im2 = canvas[y:y+h, x:x+w, :]
    
    if output_filename: 
        cv2.imwrite(output_filename, cv2.cvtColor(im2, cv2.COLOR_RGBA2BGRA))
    else:
        return canvas.astype(np.uint8)
    
def __read_image(img_path):
    img = cv2.imread(img_path, cv2.IMREAD_UNCHANGED)
    return cv2.cvtColor(img, cv2.COLOR_BGRA2RGBA)

def __extract_alpha_channel(img):
    return img[:, :, 3]

def __get_largest_contour(alpha_channel):
    # Smoothing using GaussianBlur
    smoothed = cv2.GaussianBlur(alpha_channel, (15, 15), 0)
    contours_smoothed = cv2.findContours(smoothed, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    contours_smoothed = contours_smoothed[0] if len(contours_smoothed) == 2 else contours_smoothed[1]
    big_contour_smoothed = max(contours_smoothed, key = cv2.contourArea)

    # Use the smoothed contour
    peri = cv2.arcLength(big_contour_smoothed, True)
    return cv2.approxPolyDP(big_contour_smoothed, 0.001 * peri, True)

def __draw_filled_contour_on_black_background(big_contour, shape):
    contour_img = np.zeros(shape)
    cv2.drawContours(contour_img, [big_contour], 0, 255, -1)
    return contour_img

def __apply_dilation(img):
    # TODO: This is missing in source (https://withoutbg.com/resources/creating-sticker)
    # (5, 5)
    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (30, 30))
    return cv2.morphologyEx(img, cv2.MORPH_DILATE, kernel)

def __apply_overlays(canvas, img, dilate):
    alpha = np.expand_dims(img[:, :, 3], 2)
    alpha = np.repeat(alpha, 3, 2)
    alpha = alpha / 255

    canvas[dilate == 255] = (255, 255, 255, 255)
    canvas[:, :, 0:3] = canvas[:, :, 0:3] * (1 - alpha) + alpha * img[:, :, 0:3]

    return canvas